/**
 * @file pixy2.h
 *
 * @author Jon McPhalen
 *
 * @copyright Copyright (c) Jon McPhalen, 2020. See end of file for
 * terms of use (MIT License).
 *
 * @brief Provides serial interface (up to 19.2K) to Pixy2 camera for color
 * connected component functions (i.e., color tracking). This library does
 * not require a cog for serial communications.
 *
 * @details See Pixy2 API and documentation:
 * https://docs.pixycam.com/wiki/doku.php?id=wiki:v2:start
 *
 * @note
 * This library uses non-buffered (no cog), half-duplex communications. A
 * connection failure can stop the program using the library.
 *
 * @par Memory Models
 * Use with CMM or LMM.
 *
 * @version 20 JAN 2020
 * Initial release.
 *
 * @par Help Improve this Library
 * Please submit bug reports, suggestions, and improvements to this code to
 * editor@parallax.com.
 */

#ifndef PIXY2_H
#define PIXY2_H

#if defined(__cplusplus)
extern "C" {
#endif

#include <stdint.h>


#define  BLOCK_SIZE  14                                         // block data returned in 14 bytes
#define  MAX_BLOCKS  10                                         // allow up to 10 blocks with one call
#define  BUF_SIZE     6 + (BLOCK_SIZE * MAX_BLOCKS)             // serial buffer size

#define  CCC_SIG1         0b00000001                            // color signatures
#define  CCC_SIG2         0b00000010
#define  CCC_SIG3         0b00000100
#define  CCC_SIG4         0b00001000
#define  CCC_SIG5         0b00010000
#define  CCC_SIG6         0b00100000
#define  CCC_SIG7         0b01000000
#define  CCC_ALL          0b01111111
#define  CCC_COLOR_CODES  0b10000000

#define  REQUEST_RESOLUTION     0x0C                            // get resolution (width, height) of Pixy image
#define  REQUEST_VERSION        0x0E                            // get hardware & firmware version
#define  REQUEST_BRIGHTNESS     0x10                            // set camera brightness
#define  REQUEST_SERVO          0x12                            // set pan/tilt servo positions
#define  REQUEST_LED            0x14                            // set rgb LED values
#define  REQUEST_LAMP           0x16                            // set upper and lower lamp states
#define  REQUEST_FPS            0x18                            // get frame rate
#define  REQUEST_BLOCKS         0x20                            // get signatures being tracked
#define  REQUEST_GET_RGB        0x70                            // get color information from x/y

#define  RESPONSE_RESULT        0x01                            // packet response codes
#define  RESPONSE_ERROR         0x03
#define  RESPONSE_RESOLUTION    0x0D
#define  RESPONSE_VERSION       0x0F
#define  RESPONSE_BLOCKS        0x21

#define  RESULT_OK                 0                            // result codes
#define  RESULT_ERROR             -1
#define  RESULT_BUSY              -2
#define  RESULT_CHECKSUM_ERROR    -3
#define  RESULT_TIMEOUT           -4
#define  RESULT_BUTTON_OVERRIDE   -5
#define  RESULT_PROG_CHANGING     -6


typedef struct  
{
  uint16_t signature;                                           // color signature (1..7)
  uint16_t xpos;                                                // location
  uint16_t ypos;
  uint16_t width;                                               // size
  uint16_t height;
  int16_t  angle;
  uint8_t  index;
  uint8_t  age;
} block_t;


/**
 * @brief Establish establish serial connection with Pixy
 *
 * @param rxpin Receive pin that connects to Pixy TX.
 * @param txpin Transmit pin that connects to Pixy RX.
 * @param baud Serial baud rate up to 19.2k (suggested).
 *
 * @returns 0 is successful, -1 i unable to create serial connection.
 */
int32_t pixy2_start(uint8_t rxpin, uint8_t txpin, uint16_t baud);


/**
 * @brief Request hardware and firmware version information from Pixy
 *
 * @returns 0 is successful, a negative value on communications error.
 */
int32_t pixy2_getVersion(void);


/**
 * @brief Get hardware version number after pixy_getVersion() request.
 *
 * @returns The hardware version number if successful, 0 if last request was not pixy_getVersion().
 */
uint16_t pixy2_hardware(void);


/**
 * @brief Get firmware major version number after pixy_getVersion() request.
 *
 * @returns The firmware major version number if successful, 0 if last request was not pixy_getVersion().
 */
uint8_t pixy2_fwMajor(void);


/**
 * @brief Get firmware minor version number after pixy_getVersion() request.
 *
 * @returns The firmware minor version number if successful, 0 if last request was not pixy_getVersion().
 */
uint8_t pixy2_fwMinor(void);


/**
 * @brief Get firmware build number after pixy_getVersion() request.
 *
 * @returns The firmware build number if successful, 0 if last request was not pixy_getVersion().
 */
uint16_t pixy2_fwBuild(void);


/**
 * @brief Get firmware description string after pixy_getVersion() request.
 *
 * @returns A pointer to firmware description string if successful, NULL if last request was not pixy_getVersion().
 */
char *pixy2_fwType(void);


/**
 * @brief Get image resolution details from Pixy.
 *
 * @param fWdith A pointer to unsigned 16-bit variable to hold frame width.
 * @param fHeight A pointer to unsigned 16-bit variable to hold frame height.
 *
 * @returns 0 if successful, a negative value on communications error.
 */
int32_t pixy2_getResolution(uint16_t *fWidth, uint16_t *fHeight);


/**
 * @brief Set camera brightness level.
 *
 * @param brightness Camera brightness level (0 to 255, 50 is default).
 *
 * @returns 0 if successful, a negative value on communications error.
 */
int32_t pixy2_setCameraBrightness(uint8_t brightness);


/**
 * @brief Set pan/tilt servo positions.
 *
 * @param s0 Pan position (0 to 511).
 * @param s0 Tilt position (0 to 511).
 *
 * @returns 0 if successful, a negative value on communications error.
 */
int32_t pixy2_setServos(uint16_t s0, uint16_t s1);


/**
 * @brief Set RGB LED brightness levels.
 *
 * @param r Red level (0 to 255).
 * @param g Green level (0 to 255).
 * @param b Blue level (0 to 255).
 *
 * @returns 0 if successful, a negative value on communications error.
 */
int32_t pixy2_setLED(uint8_t r, uint8_t g, uint8_t b);


/**
 * @brief Set upper and lower lamp states (off or on).
 *
 * @param upper Set state (0 or 1) of upper lamp LEDs.
 * @param lower Set state (0 or 1) of lower lamp (RGB) LEDs.
 *
 * @returns 0 if successful, a negative value on communications error.
 */
int32_t pixy2_setLamp(uint8_t upper, uint8_t lower);


/**
 * @brief Get frame rate from Pixy camera.
 *
 * @returns frame rate if successful, a negative value on communications error.
 */
int32_t pixy2_getFPS(void);


/**
 * @brief Request blocks (tracking data) for specified signatures.
 * 
 * @note If using color codes, an octal value returned; digits in this
 * value represent the color signatures that make up the the color code.
 *
 * @param sigmap Bitmap of desired signature(s).
 * @param maxBlocks The maximum number of blocks that to be returned.
 *
 * @returns number of blocks in response packet, a negative value on communications error.
 */
uint8_t pixy2_getBlocks(uint8_t sigmap, uint8_t maxBlocks);


/**
 * @brief Move block data from recieve buffer to block_t variable.
 *
 * @param i Index (0..blockCount-1) of desired block.
 * @param *block A pointer to block_t struct to hold the block (tracking) values.
 *
 * @returns 0 if successful, -1 if bad block index or no block data in buffer.
 */
int32_t pixy2_extractBlock(uint8_t i, uint16_t *block);



/**
 * @brief Get RGB color information for desired x/y coordinate.
 *
 * @param x Horizontal center of position to scan.
 * @param y Vertical cetner of position to scan.
 * @param saturate If set (1), brightest color component will be set to 0xFF
 * and others will be scaled proportionately.
 *
 * @returns 0 if successful, a negative value on communications error.
 */
int32_t pixy2_getRGB(uint16_t x, uint16_t y, uint8_t saturate);




// utilities used only in pixy2.c

void      clearBuffer(void);                                    // empty rx buffer
uint8_t   recvPacket(void);                                     // wait for and capture Pixy response
int32_t   validateChecksum();                                   // valid checksum of packet in buffer
uint16_t  extract16(uint8_t i);                                 // extract 16-bit value from buffer
uint32_t  extract32(uint8_t i);                                 // extract 32-bit value from buffer
uint32_t  extractLE(uint8_t i, uint8_t n);                      // extract value from buffer


#if defined(__cplusplus)
}
#endif
/* __cplusplus */

#endif
/* PIXY2_H */


/**
 * TERMS OF USE: MIT License
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
 * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 * DEALINGS IN THE SOFTWARE.
 */
